-- This file is required file for icekit.exe; icekit.exe loads this
-- file upon startup. see aice_util.txt for function descriptions

require("aice");
require("bit");
require("bit32");

sleep = aice.sleep

function printf(format, ...)
	io.write(string.format(format, ...))

	-- io.output():setvbuf("line") doesn't seem to work reliably.
	if string.match(format, "\n") ~= nil then
		io.flush()
	end
end
function pause(format, ...)
	io.write(string.format(format, ...))
	io.flush()
	return io.read()
end

function sprintf(format, ...)
	return string.format(format, ...)
end

function exit(...)
	os.exit(...)
end

function eq(w1, w2)
	return aice.word_eq(w1, w2)
end

function neq(w1, w2)
	return not aice.word_eq(w1, w2)
end

function tohex(...)
	return aice.tohex(...)
end

function fromhex(str)
	str = string.gsub(str, "^0x", "")
	return tonumber(str, 16)
end

-- portability provision layer for LUA 5.2
bit_not = bit32.bnot
bit_and = bit32.band
bit_or = bit32.bor
bit_xor = bit32.bxor
bit_rshift = bit32.rshift
bit_arshift = bit32.arshift
bit_lshift = bit32.lshift
byte_reverse = aice.byte_reverse

function bit_test(value, mask)
	local val = bit_and(value, mask)
	if (val == 0) then
		return false
	else
		return true
	end
end

bit_set = bit32.bor
function bit_clear(value, mask)
	local val = bit_or(value, mask)
	return bit_xor(val, mask)
end

function bit_extract(value, hi, lo)
	local v = bit_lshift(value,31-hi)
	return bit_rshift(v,lo+31-hi)
end

function bool2int(bool)
	if bool then
		return 1
	else
		return 0
	end
end
function is_zero(int)
	if (int == 0) then
		return true
	else
		return false
	end
end
function not_zero(int)
	if (int ~= 0) then
		return true
	else
		return false
	end
end

function def_edmsr_index(major, minor, ext)
	x = (minor % 16) * 8 + (ext%8)
	return x
end
function def_sys_reg_index(major, minor, ext)
	x = (major%8) * 128 + (minor % 16) * 8 + (ext%8)
	return x
end

_G['sdp_fast_mode'] = true

_edmsr = {}
_misc = {}
_actrl = {}
_sys = {}

for i = 0,7 do
_edmsr["BPC"   .. i] = def_edmsr_index(3, 0, i)
_edmsr["BPA"   .. i] = def_edmsr_index(3, 1, i)
_edmsr["BPAM"  .. i] = def_edmsr_index(3, 2, i)
_edmsr["BPV"   .. i] = def_edmsr_index(3, 3, i)
_edmsr["BPCID" .. i] = def_edmsr_index(3, 4, i)
end

_edmsr["EDM_CFG"] = def_edmsr_index(3, 5, 0)
_edmsr["EDMSW"] = def_edmsr_index(3, 6, 0)
_edmsr["EDM_CTL"] = def_edmsr_index(3, 7, 0)
_edmsr["EDM_DTR"] = def_edmsr_index(3, 8, 0)
_edmsr["BPMTC"] = def_edmsr_index(3, 9, 0)
_edmsr["DIMBR"] = def_edmsr_index(3, 10, 0)
_edmsr["TECR0"] = def_edmsr_index(3, 14, 0)
_edmsr["TECR1"] = def_edmsr_index(3, 14, 1)
_edmsr["DTR"] = def_edmsr_index(3, 8, 0) -- alias

_misc["DIMIR"] = 0
_misc["SBAR"]  = 1
_misc["EDM_CMDR"] = 2
_misc["DBGER"] = 3
_misc["ACC_CTL"] = 4
_misc["GPORT0"] = 6
_misc["GPORT1"] = 7

_sys["CPU_VER"] = def_sys_reg_index(0, 0, 0)
_sys["ICM_CFG"] = def_sys_reg_index(0, 1, 0)
_sys["DCM_CFG"] = def_sys_reg_index(0, 2, 0)
_sys["MMU_CFG"] = def_sys_reg_index(0, 3, 0)
_sys["MSC_CFG"] = def_sys_reg_index(0, 4, 0)
_sys["CORE_ID"] = def_sys_reg_index(0, 0, 1)
_sys["FUCOP_EXIST"] = def_sys_reg_index(0, 5, 0)
_sys["PSW"] = def_sys_reg_index(1, 0, 0)
_sys["IPSW"] = def_sys_reg_index(1, 0, 1)
_sys["P_IPSW"] = def_sys_reg_index(1, 0, 2)
_sys["IVB"] = def_sys_reg_index(1, 1, 1)
_sys["EVA"] = def_sys_reg_index(1, 2, 1)
_sys["P_EVA"] = def_sys_reg_index(1, 2, 2)
_sys["ITYPE"] = def_sys_reg_index(1, 3, 1)
_sys["P_ITYPE"] = def_sys_reg_index(1, 3, 2)
_sys["MERR"] = def_sys_reg_index(1, 4, 1)
_sys["IPC"] = def_sys_reg_index(1, 5, 1)
_sys["P_IPC"] = def_sys_reg_index(1, 5, 2)
_sys["OIPC"] = def_sys_reg_index(1, 5, 3)
_sys["P_P0"] = def_sys_reg_index(1, 6, 2)
_sys["P_P1"] = def_sys_reg_index(1, 7, 2)
_sys["INT_MASK"] = def_sys_reg_index(1, 8, 0)
_sys["INT_PEND"] = def_sys_reg_index(1, 9, 0)
_sys["MMU_CTL"] = def_sys_reg_index(2, 0, 0)
_sys["L1_PPTB"] = def_sys_reg_index(2, 1, 0)
_sys["TLB_VPN"] = def_sys_reg_index(2, 2, 0)
_sys["TLB_DATA"] = def_sys_reg_index(2, 3, 0)
_sys["TLB_MISC"] = def_sys_reg_index(2, 4, 0)
_sys["VLPT_IDX"] = def_sys_reg_index(2, 5, 0)
_sys["ILMB"] = def_sys_reg_index(2, 6, 0)
_sys["DLMB"] = def_sys_reg_index(2, 7, 0)
_sys["CACHE_CTL"] = def_sys_reg_index(2, 8, 0)
_sys["HSMP_SADDR"] = def_sys_reg_index(2, 9, 0)
_sys["HSMP_EADDR"] = def_sys_reg_index(2, 9, 1)
for i = 0,7 do
_sys["BPC"   .. i] = def_sys_reg_index(3, 0, i)
_sys["BPA"   .. i] = def_sys_reg_index(3, 1, i)
_sys["BPAM"  .. i] = def_sys_reg_index(3, 2, i)
_sys["BPV"   .. i] = def_sys_reg_index(3, 3, i)
_sys["BPCID" .. i] = def_sys_reg_index(3, 4, i)
end

_sys["EDM_CFG"] = def_sys_reg_index(3, 5, 0)
_sys["EDMSW"] = def_sys_reg_index(3, 6, 0)
_sys["EDM_CTL"] = def_sys_reg_index(3, 7, 0)
_sys["EDM_DTR"] = def_sys_reg_index(3, 8, 0)
_sys["BPMTC"] = def_sys_reg_index(3, 9, 0)
_sys["DIMBR"] = def_sys_reg_index(3, 10, 0)
_sys["TECR0"] = def_sys_reg_index(3, 14, 0)
_sys["TECR1"] = def_sys_reg_index(3, 14, 1)
_sys["DTR"] = def_sys_reg_index(3, 8, 0) -- alias

_sys["PFMC0"] = def_sys_reg_index(4, 0, 0)
_sys["PFMC1"] = def_sys_reg_index(4, 0, 1)
_sys["PFMC2"] = def_sys_reg_index(4, 0, 2)
_sys["PFM_CTL"] = def_sys_reg_index(4, 1, 0)

_sys["DMA_CFG"] = def_sys_reg_index(5, 0, 0)
_sys["DMA_GCSW"] = def_sys_reg_index(5, 1, 0)
_sys["DMA_CHNSEL"] = def_sys_reg_index(5, 2, 0)
_sys["DMA_ACT"] = def_sys_reg_index(5, 3, 0)
_sys["DMA_SETUP"] = def_sys_reg_index(5, 4, 0)
_sys["DMA_ISADDR"] = def_sys_reg_index(5, 5, 0)
_sys["DMA_ESADDR"] = def_sys_reg_index(5, 6, 0)
_sys["DMA_TCNT"] = def_sys_reg_index(5, 7, 0)
_sys["DMA_STATUS"] = def_sys_reg_index(5, 8, 0)
_sys["DMA_2DSET"] = def_sys_reg_index(5, 9, 0)
_sys["DMA_2DSCTL"] = def_sys_reg_index(5, 9, 1)

_sys["PRUSR_ACC_CTL"] = def_sys_reg_index(4, 4, 0)
_sys["FUCOP_CTL"] = def_sys_reg_index(4, 5, 0)

_sys["SDZ_CTL"] = def_sys_reg_index(2, 15, 0)
_sys["MISC_CTL"] = def_sys_reg_index(2, 15, 1)

-- for read
_actrl["ICE_STATUS"] = 0x0
_actrl["HW_VER"] = 0x1
_actrl["FPGA_VER"] = 0x2
_actrl["FW_VER"] = 0x3
_actrl["PIN_STATUS"] = 0x4

_actrl["TCK_CTRL"] = 0x0
_actrl["JTAG_CTRL"] = 0x1
_actrl["CLEAR_TIMEOUT"] = 0x2

_actrl["BATCH_CTRL"] = 0x20
_actrl["BATCH_ITERATION"] = 0x21
_actrl["BATCH_DIM_SIZE"] = 0x22
_actrl["BATCH_BUF_INFO"] = 0x22
_actrl["BATCH_STATUS"] = 0x23
_actrl["BATCH_BUF0_CTRL"] = 0x30
_actrl["BATCH_BUF0_STATE"] = 0x31

function s(name)
	return _edmsr[string.upper(name)];
end

function m(name)
	return _misc[string.upper(name)];
end

function ac(name)
	return _actrl[string.upper(name)]
end

TCK_48MHz = 8
TCK_24MHz = 9
TCK_12MHz = 10
TCK_6MHz  = 11
TCK_3MHz  = 12
TCK_3000KHz = 12 -- alias
TCK_1500KHz = 13
TCK_750KHz = 14
TCK_375KHz = 15

TCK_30MHz   = 0
TCK_15MHz   = 1
TCK_7MHz    = 2
TCK_3750KHz = 3
TCK_1875KHz = 4
TCK_937KHz  = 5
TCK_468KHz  = 6
TCK_234KHz  = 7

-- tck divider to frequency
function tckdiv2freq(div)
	local freq
	if (bit_and(div, 0x8)~= 0) then
		div = bit_and(div, 0x7)
		freq = 48/(2^div)
	else
		freq = 30/(2^div)
	end
	return freq
end

AICE_CTRL4_WRITE_PINS_SUPPORT = 0x40000000
AICE_CTRL4_NODBGI_MASK = 0x4
AICE_CTRL4_FASTMODE_MASK = 0x2

L1D_IX_INVAL	= 0x00	-- 0b00000
L1D_IX_WB   	= 0x01	-- 0b00001
L1D_IX_WBINVAL	= 0x02	-- 0b00010
L1D_IX_RTAG	= 0x03	-- 0b00011
L1D_IX_RWD	= 0x04	-- 0b00100
L1D_IX_WTAG	= 0x05	-- 0b00101
L1D_IX_WWD	= 0x06	-- 0b00110
L1D_INVALALL	= 0x07	-- 0b00111

L1D_VA_INVAL	= 0x08	-- 0b01000
L1D_VA_WB   	= 0x09	-- 0b01001
L1D_VA_WBINVAL	= 0x0a	-- 0b01010
L1D_VA_FILLCK	= 0x0b	-- 0b01011
L1D_VA_ULCK	= 0x0c	-- 0b01100

L1I_IX_INVAL	= 0x10	-- 0b10000
L1I_IX_RTAG	= 0x13	-- 0b10011
L1I_IX_RWD	= 0x14	-- 0b10100
L1I_IX_WTAG	= 0x15	-- 0b10101
L1I_IX_WWD	= 0x16	-- 0b10110

L1I_VA_INVAL	= 0x18	-- 0b11000
L1I_VA_FILLCK	= 0x1b	-- 0b11011
L1I_VA_ULCK	= 0x1c	-- 0b11100

function hex_dump(buf)
   for byte=1, #buf, 16 do
      local chunk = buf:sub(byte, byte+15)
      printf('%08X  ',byte-1)
      chunk:gsub('.', function (c) printf('%02X ',string.byte(c)) end)
      io.write(string.rep(' ',3*(16-#chunk)))
      io.write(' ',chunk:gsub('%c','.'),"\n") 
   end
end
function get_aice_reg(name)
	return aice.cmd_read_ctrl(ac(name))
end
function set_aice_reg(name, value)
	return aice.cmd_write_ctrl(ac(name), value)
end
function is_aice_box()
	local hw_ver
	hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end
	local aice_box = bit_rshift(hw_ver, 16) == 0
	return aice_box
end
function is_aice_mcu()
	local hw_ver
	hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end
	local aice_mcu = bit_rshift(hw_ver, 16) == 1
	return aice_mcu
end
function is_aice_mini()
	local hw_ver
	hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end
	local aice_mini = bit_rshift(hw_ver, 16) == 2
	return aice_mini
end
function get_jtag_interface()
	local ctrl4 = aice.cmd_read_ctrl(0x4)	-- JTAG pin status
	ctrl4 = bit_and(ctrl4, 0x3)
	if (ctrl4 == 0x1) then
		return "2 wire JTAG (Basic Mode)"
	elseif (ctrl4 == 0x3) then
		return "2 wire JTAG (Fast Mode)"
	else
		return "5 wire JTAG"
	end
end
function has_no_dbgi_pin()
	local ctrl4 = aice.cmd_read_ctrl(0x4)
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end
	return bit_test(ctrl4,AICE_CTRL4_NODBGI_MASK)
end
function is_v2_arch()
	local edm_cfg = dbg_get_edm_reg("EDM_CFG")
	local ARCH = bit_rshift(edm_cfg, 28)
	return (ARCH == 0)
end
function get_edm_version()
	local edm_cfg = dbg_get_edm_reg("EDM_CFG")
	local VER     = bit_extract(edm_cfg,31,16)
	return VER
end
function is_edm_support_dalm()
	local edm_cfg = dbg_get_edm_reg("EDM_CFG")
	return bit_test(edm_cfg, 0x10)
end
function is_edm_support_hw_debug_on_reset()
	local	cpu_ver = dbg_get_sys_reg("CPU_VER")
	local	cpu_id = bit_extract(cpu_ver, 31, 16)
	return ((get_edm_version() >= 0x51) and (cpu_id ~= 0x0c1c))
end
function is_edm_misc_in_tck_domain()
	local	cpu_ver = dbg_get_sys_reg("CPU_VER")
	local	cpu_id = bit_extract(cpu_ver, 31, 16)
	return ((get_edm_version() >= 0x51) and (cpu_id ~= 0x0c1c))
end
function reset_and_enter_dbg()
	local num_of_devs = get_number_of_jtag_devices()
	if (num_of_devs == 1) then
		select_first_andes_core()
		-- FOR AG101
		aice.cmd_sys_reset()		-- avoid setup_dim fail, in case core is standby
		aice.sleep(700)			-- wait system reset done
		dbg_setup_dim(0xfff00000)
		aice.cmd_restart()
		if (aice_cmd_timeout) then
			aice.cmd_sys_reset()
			aice.sleep(700)		-- wait system reset done
			dbg_setup_dim(0xfff00000)
			dbg_set_edm_reg("EDM_CMDR", 0x0) -- trigger the debug exception
		end
	elseif (num_of_devs > 1) then
		-- FOR AG102
		printf("Debug on Reset for AG102...")
		aice.cmd_write_ctrl(0xd, 0x02000001)	-- sys_rst_n = 0, delay 0x200 ms
		aice.cmd_write_ctrl(0xd, 0x02000002)	-- sys_rst_n = 1, delay 0x200 ms

		for i=0,num_of_devs-1 do
			aice.cmd_t_write_edmsr(i, s("DIMBR"), 0xfff00000)
			aice.cmd_t_write_edmsr(i, s("EDM_CTL"), 0x8000000c)
			aice.cmd_t_write_misc(i, m("EDM_CMDR"), 0x0)
		end
		printf("ok\n")
	else
		error(string.format("Error: Can not detect any devices\n"))
	end
end
function get_number_of_jtag_devices()
	local jtag_id_table = {}
	local number = 0
	jtag_id_table = aice.cmd_scan_chain();
	number = 0
	for i,v in ipairs(jtag_id_table) do
		number = number+1
	end
	return number
end
function select_first_andes_core()
	local ANDES_JTAG_ID = 0x1000063d
	local ANDES_JTAG_ID_MASK = 0xFF000FFF
	local jtag_id_table = {}
	local target
	local v = aice.set_cmd_verbose(0)

	jtag_id_table = aice.cmd_scan_chain();

	if (aice_cmd_timeout) then
		error("[SCAN_CHAIN] can not access the JTAG scan chain\n", 2)
	end

	local i, idcode
	for i,idcode in ipairs(jtag_id_table) do
		local base_idcode = bit_and(idcode, ANDES_JTAG_ID_MASK)
		if (base_idcode == ANDES_JTAG_ID) then
			target = i
			printf("JTAG_ID: %08x selected\n", idcode)
			break
		end
	end

	if (target == nil) then
		error("can not find AndesCore in the JTAG scan chain\n", 2)
	end

	aice.cmd_select_target(target)

	if (aice_cmd_timeout) then
		error("[SELECT_TARGET] can not select AndesCore for debugging\n", 2)
	end

	aice.set_cmd_verbose(v)
	return target
end

function get_aice_version()
	local hw_ver
	local fpga_ver
	local fw_ver
	local dev_type
	local aice_mcu
	local aice_dev_suffix
	local aice_mini
	
	local v = aice.set_cmd_verbose(0)
	hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end
	dev_type = bit_rshift(hw_ver, 16)
	aice_mcu = (dev_type == 1)
	aice_mini = (dev_type == 2)
	if (aice_mcu) then
		aice_dev_suffix = "_mcu"
	elseif (aice_mini) then
		aice_dev_suffix = "_mini"
	else
		aice_dev_suffix = ""
	end
	hw_ver = bit_and(hw_ver, 0xffff)
	if (aice_cmd_timeout) then
		return 0, 0, 0
	end

	fpga_ver = aice.cmd_read_ctrl(0x2) -- fpga prom version
	fw_ver = aice.cmd_read_ctrl(0x3) -- firmware version

	if (aice_cmd_timeout) then
		error("USB connection or icebox broken?!", 2)
	end

	_G['aice_mcu'] = aice_mcu -- ugly hack to provide global aice_mcu flag
	_G['aice_mini'] = aice_mini -- ugly hack to provide global aice_mini flag
	printf("aice%s version: %s.%s.%s\n", aice_dev_suffix, hw_ver, fw_ver, fpga_ver)
	aice.set_cmd_verbose(v)
	return hw_ver, fw_ver, fpga_ver, aice_mcu, aice_mini
end

function enable_dbgi_pin()
	local hw_ver = aice.cmd_read_ctrl(0x1) -- hardware version
	local dev_type = bit_rshift(hw_ver, 16)
	local aice_mcu = dev_type == 1
	local aice_mini = dev_type == 2
	if (aice_mcu or aice_mini) then
		return
	end
	local aice_ctrl4 = aice.cmd_read_ctrl(0x4)
	if (bit_test(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)) then
		-- clear AICE_CTRL4_NODBGI_MASK
		aice_ctrl4 = bit_or(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)
		aice_ctrl4 = bit_xor(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)
		aice.cmd_write_ctrl(0x4, aice_ctrl4)
	end
end

function disable_sdp_fast_mode()
	_G['sdp_fast_mode'] = false	
	local ctrl4 = aice.cmd_read_ctrl(0x4)
	ctrl4 = bit_clear(ctrl4, AICE_CTRL4_FASTMODE_MASK)
	aice.cmd_write_ctrl(0x4, ctrl4)
	ctrl4 = aice.cmd_read_ctrl(0x4)
	if (bit_test(ctrl4, AICE_CTRL4_FASTMODE_MASK)) then
		error(string.format("ERROR: can not disable fast mode\n"))
	end
end

function enable_sdp_fast_mode()
	_G['sdp_fast_mode'] = true
	local ctrl4 = aice.cmd_read_ctrl(0x4)
	ctrl4 = bit_or(ctrl4, AICE_CTRL4_FASTMODE_MASK)
	aice.cmd_write_ctrl(0x4, ctrl4)
	ctrl4 = aice.cmd_read_ctrl(0x4)
	if (not bit_test(ctrl4, AICE_CTRL4_FASTMODE_MASK)) then
		error("ERROR: The device does not support fast mode")
	end
end

function dbg_set_dbger(value)
	local RETRY_COUNT = 2
	local v = aice.set_cmd_verbose(0)
	local reg, done, retry

	retry = 0;
	done = false
	while not done do
		reg = aice.cmd_write_misc(m('DBGER'), value)
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error("set DBGER timeout", 2);
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
	return reg
end

function caller_info()
	local info = debug.getinfo(3, "Snl")
	local short_src = "lua"
	local currentline = 0
	local caller = "unknown_func"

	if (info ~= nil) then
		short_src = info.short_src
		caller = info.name
		currentline = info.currentline
		if (short_src ~= nil) then
			return string.format("%s() at %s:%d ", caller, short_src, currentline)
		elseif (caller ~= nil) then
			return string.format("%s() ", caller)
		else
			return "y"
		end
	else
		return "x"
	end
end

function dbg_get_dbger()
	local RETRY_COUNT = 2
	local v = aice.set_cmd_verbose(0)
	local reg, done, retry

	retry = 0
	done = false
	while not done do
		reg = aice.cmd_read_misc(m('DBGER'))
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error("read DBGER timeout", 2);
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
	return reg
end

function dbg_get_dtr()
	local v
	local reg
	v = aice.set_cmd_verbose(0)
	reg = aice.cmd_read_dtr()
	aice.set_cmd_verbose(v)

	return reg
end


function dbg_set_dtr(value)
	local v
	v = aice.set_cmd_verbose(0)
	aice.cmd_write_dtr(value)
	aice.set_cmd_verbose(v)
end

function dbg_set_edm_reg(name, value)
	local edmsr_index = s(name)
	local misc_index = m(name)
	local is_dtr
	local v
	v = aice.set_cmd_verbose(0)
	if (name == "EDM_DTR" or name == "DTR") then
		is_dtr = true
	else
		is_dtr = false
	end

	local RETRY_COUNT = 2
	local reg, done, retry

	done = false
	retry = 0
	while not done do
		if (is_dtr) then
			aice.cmd_write_dtr(value)
		elseif (misc_index ~= nil) then
			aice.cmd_write_misc(misc_index, value)
		elseif (edmsr_index ~= nil) then
			aice.cmd_write_edmsr(edmsr_index, value)
		else
			error(name .. " is not a recognized debug register", 2)
		end
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error(string.format("ERROR: TIMEOUT: failed to write %08x to %s\n", value, name), 2)
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
end

function dbg_get_edm_reg(name, value)
	local edmsr_index = s(name)
	local misc_index = m(name)
	local is_dtr;
	local v
	v = aice.set_cmd_verbose(0)

	if (name == "EDM_DTR" or name == "DTR") then
		is_dtr = true
	else
		is_dtr = false
	end

	local RETRY_COUNT = 2
	local reg, done, retry

	retry = 0
	done = false
	while not done do
		if (is_dtr) then
			reg = aice.cmd_read_dtr()
		elseif (misc_index ~= nil) then
			reg = aice.cmd_read_misc(misc_index)
		elseif (edmsr_index ~= nil) then
			reg = aice.cmd_read_edmsr(edmsr_index)
		else
			error(name .. " is not a recognized debug register", 2)
		end
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error(string.format("TIMEOUT reading %s", name), 2)
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_cmd_verbose(v)
	return reg
end

-- setup DIM. optional argument: dim_base_address. default = 0xfff00000
function dbg_setup_dim(...)
	local dim_base = ...
	local v = aice.set_cmd_verbose(0)
	if (dim_base == nil) then
		dim_base = 0xfff00000
	end

	dbg_set_edm_reg("DIMBR", dim_base)

	local edm_ctl = dbg_get_edm_reg("EDM_CTL")
	edm_ctl = bit_or(edm_ctl, 0x80000000)
	dbg_set_edm_reg("EDM_CTL", edm_ctl)

	dbg_set_dbger(0x2) -- clear DPED
	aice.set_cmd_verbose(v)
end

function dbg_wait_dex(...)
	local RETRY_COUNT = ...
	-- pulling until DBGER.DEX set
	local retry = 0
	if RETRY_COUNT == nil then
		RETRY_COUNT = 100
	end

	local dbger = 0
	while dbger == 0 do
		dbger = dbg_get_dbger()
		dbger = bit_and(dbger, 0x1)
		if (aice_cmd_timeout) then
			error("failed to get DBGER status", 2);
		end
		retry = retry + 1
		if (retry > RETRY_COUNT) then
			return false
		end
		aice.sleep(1)
	end
	return true
end

-- when DEX=0, CPU might still be in DEX mode. this function
-- test this scenario and it must be run with DEX==0
function dbg_test_dex_mode()
	local v = aice.set_cmd_verbose(0)

	dbg_set_dbger(0x3) -- clear DPED & DEX
	
	-- test if CPU is already in DEX mode
-- 64000008              nop
-- 64000008              nop
-- 64000008              nop
-- 48fffffa              j 1b
	local w = {0x64000008, 0x64000008, 0x64000008, 0x48fffffa}
	w = reverse_endian(w)
	dbg_set_edm_reg("DIMIR", 0)
	aice.cmd_write_dim(w[1], w[2], w[3], w[4])
	if (aice_cmd_timeout) then
		aice.cmd_clear_timeout()
		return false
	end

	local RETRY_COUNT = 2
	local reg, done, retry
	retry = 0
	done = false
	while not done do
		aice.cmd_execute()
		if (aice_cmd_timeout) then
			retry = retry + 1
			aice.cmd_clear_timeout()
			if (retry >= RETRY_COUNT) then
				return false
			end
		else
			done = true
		end
	end

	RETRY_COUNT = 2
	retry = 0
	done = false
	while not done do
		reg = dbg_get_dbger()
		if (aice_cmd_timeout) then
			aice.cmd_clear_timeout()
			return false
		end
		reg = bit_and(reg, 0x2)
		done = (reg == 0x2)
		retry = retry + 1
		if (retry > 3) then
			aice.sleep(1)
		end
		if (not done and (retry > RETRY_COUNT)) then
			return false -- not in DEX mode
		end
	end

	-- clear DPEX bit (write one to clear)
	dbg_set_dbger(0x2)

	aice.set_cmd_verbose(v)
	return true	-- but DEX = 0
end

-- optional argument use_dbgi_pin; when set, use dbgi_n to trigger the debug exception
-- otherwise, use the DBGI_CMD.
-- dbg_break will wait until DEX is set
function dbg_break(...)
	local use_dbgi_pin = ...
	local v = aice.set_cmd_verbose(0)

	if (use_dbgi_pin == nil) then
		use_dbgi_pin = false
	end

	local dbger = dbg_get_dbger()
	dbger = bit_and(dbger, 0x1)
	if dbger ~= 0 then
		aice.set_cmd_verbose(v)
		return
	end

	local in_dex_mode = dbg_test_dex_mode()
	if in_dex_mode then
		dbg_continue() -- so that EDM responds to the break commands below
	end
	
	if (use_dbgi_pin) then
		aice.cmd_assert_dbgi()
		if (aice_cmd_timeout) then
			error("TIMEOUT asserting dbgi_n pin", 2)
		end
	else
		dbg_set_edm_reg("EDM_CMDR", 0x0) -- trigger the debug exception
	end

	if not dbg_wait_dex(100) then
		error("failed to trigger debug exception", 2);
	end

	aice.set_cmd_verbose(v)
end

function reverse_endian(table)
	local rt = {}
	for i,v in ipairs(table) do
		rt[i] = byte_reverse(v)
	end
	return rt
end

-- w is an array of 4 words. the words are instructions in big endian format
-- as shown by objdump output.
--
-- CPU must be in the debug state, DIMBR must be setup to use this function
function dbg_execute(w)
	local w = reverse_endian(w)
	
	dbg_set_edm_reg("DIMIR", 0)

	aice.cmd_write_dim(w[1], w[2], w[3], w[4])

	local RETRY_COUNT = 2
	local reg, done, retry
	retry = 0
	done = false
	while not done do
		aice.cmd_execute()
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				local caller = caller_info()
				local stack_level = 2
				if string.match(caller, "aice_util.lua") ~= nil then
					stack_level = 3
				end
				error("failed to resume execution", stack_level);
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end

	RETRY_COUNT = 100
	retry = 0
	done = false
	while not done do
		reg = dbg_get_dbger()
		if (aice_cmd_timeout) then
			error("failed to resume execution", 2);
		end
		reg = bit_and(reg, 0x2)
		done = (reg == 0x2)
		retry = retry + 1
		if (retry > 3) then
			aice.sleep(1)
		end
		if (not done and (retry > RETRY_COUNT)) then
			local caller = caller_info()
			local stack_level = 2
			if string.match(caller, "aice_util.lua") ~= nil then
				stack_level = 3
			end
			error(string.format("execution not done"), stack_level)
		end
	end

	-- clear DPEX bit (write one to clear)
	dbg_set_dbger(0x2)
	if (aice_cmd_timeout) then
		error("failed to resume execution", 2);
	end
end

-- This function requires DIM to be setup
-- dbg_get_pc()
--     get PC when the debug exception happened
--
--               1:      ! get IPC
-- 6402A402              mfsr    $r0, $IPC
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              dsb
-- 48fffffa              j 1b
--
--               1:      ! get OIPC
-- 6402AC02              mfsr    $r0, $OIPC
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_get_pc()
	local v = aice.set_cmd_verbose(0)
	local reg
	local r0 = dbg_get_gpr(0)
	if (is_v2_arch()) then
		dbg_execute({0x6402A402, 0x64070003, 0x64000008, 0x48fffffa})
		reg = dbg_get_dtr()
		if (aice_cmd_timeout) then
			error("failed to get ipc", 2)
		end
	else
		dbg_execute({0x6402AC02, 0x64070003, 0x64000008, 0x48fffffa})
		reg = dbg_get_dtr()
		if (aice_cmd_timeout) then
			error("failed to get oipc", 2)
		end
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- This function requires DIM to be setup
-- dbg_set_pc()
--     set IPC
--
--               1:      ! set IPC
-- 64070002              mfsr    $r0, $EDM_DTR
-- 6402A403              mtsr    $r0, $IPC
-- 64000008              dsb
-- 48fffffa              j 1b
--
--     set OIPC
--               1:      ! set IPC
-- 64070002              mfsr    $r0, $EDM_DTR
-- 6402AC03              mtsr    $r0, $OIPC
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_set_pc(value)
	local v = aice.set_cmd_verbose(0)
	local reg
	local r0 = dbg_get_gpr(0)

	if (is_v2_arch()) then
		dbg_set_dtr(value)
		dbg_execute({0x64070002, 0x6402A403, 0x64000008, 0x48fffffa})
		if (aice_cmd_timeout) then
			error("failed to set ipc", 2)
		end
	else
		dbg_set_dtr(value)
		dbg_execute({0x64070002, 0x6402AC03, 0x64000008, 0x48fffffa})
		if (aice_cmd_timeout) then
			error("failed to set oipc", 2)
		end
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- This function requires DIM to be setup
-- dbg_get_sys_reg()
--     get system register
--
--               1:      ! get system registers
-- 64000002              mfsr    $r0, $XXX
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_get_sys_reg(sys_reg_name)
	local v = aice.set_cmd_verbose(0)
	local reg
	local w = {0x64000002, 0x64070003, 0x64000008, 0x48fffffa}
	local sys_reg_index = _sys[sys_reg_name]
	if sys_reg_index == nil then
		error(string.format("invalid register name: %s\n", sys_reg_name), 2)
	end

	local r0 = dbg_get_gpr(0)
	w[1] = bit_or(w[1], bit_lshift(sys_reg_index, 10))
	dbg_execute(w)
	reg = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error(string.format("failed to get %s", sys_reg_name), 2)
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- This function requires DIM to be setup
-- dbg_set_sys_reg()
--     get system register
--
--               1:      ! set system registers
-- 64070002              mfsr    $r0, $EDM_DTR
-- 64000003              mtsr    $r0, $XXX
-- 64000008              dsb
-- 48fffffa              j 1b
function dbg_set_sys_reg(sys_reg_name, value)
	local v = aice.set_cmd_verbose(0)
	local r0 = dbg_get_gpr(0)

	dbg_set_dtr(value)

	local w = {0x64070002, 0x64000003, 0x64000008, 0x48fffffa}
	local sys_reg_index = _sys[sys_reg_name]
	if sys_reg_index == nil then
		error(string.format("invalid register name: %s\n", sys_reg_name), 2)
	end

	w[2] = bit_or(w[2], bit_lshift(sys_reg_index, 10))
	dbg_execute(w)
	if (aice_cmd_timeout) then
		error(string.format("failed to get %s", sys_reg_name), 2)
	end
	dbg_set_gpr(0, r0)
	aice.set_cmd_verbose(v)
	return reg
end

-- dbg_continue()
--     let target resume execution and leaves debug mode
--
-- 40000009              nop
-- 40000009              nop
-- 40000009              nop
-- 64000004              iret
function dbg_continue()
	local v = aice.set_cmd_verbose(0)
	local reg

	dbg_set_dbger(0x1) -- clear DEX

	dbg_execute({0x40000009, 0x40000009, 0x40000009, 0x64000004})
	aice.set_cmd_verbose(v)
end

-- reach_max_intl()
--	detect whether the max interrupt level is reached or not 
function reach_max_intl()
	local msc_cfg, psw
	local intlc, intl, max_intl

	msc_cfg = dbg_get_sys_reg('MSC_CFG')
	intlc = bit_test(msc_cfg, 0x1000)
	psw = dbg_get_sys_reg('PSW')
	intl = bit_rshift(bit_and(psw, 0x6), 1)

	if (intl>=3) then
		max_intl = true
	elseif (intlc and (intl>=2)) then
		max_intl = true
	else
		max_intl = false
	end
	return max_intl
end

-- dbg_step()
--     single step one instruction
function dbg_step()
	local v = aice.set_cmd_verbose(0)
	local psw, ipsw

	if (is_v2_arch() and not reach_max_intl()) then
		ipsw = dbg_get_sys_reg('IPSW')
		ipsw = bit_or(ipsw, 0xc00)  -- enable HSS, set DEX
		ipsw = bit_xor(ipsw, 0x400) -- enable HSS, clear DEX
		dbg_set_sys_reg('IPSW', ipsw)
	else
		psw = dbg_get_sys_reg('PSW')
		psw = bit_or(psw, 0x800)        -- enable HSS
		dbg_set_sys_reg('PSW', psw)
	end

	dbg_continue()
	if not dbg_wait_dex(200) then
		error("control is not returned to debugger after single stepping\n", 2);
	end

	if (is_v2_arch() and not reach_max_intl()) then
		ipsw = dbg_get_sys_reg('IPSW')
		ipsw = bit_or(ipsw, 0x800)
		ipsw = bit_xor(ipsw, 0x800) -- disable HSS
	
		psw = dbg_get_sys_reg('PSW')
		psw = bit_or(psw, 0x800)
		psw = bit_xor(psw, 0x800) -- disable HSS
		dbg_set_sys_reg('IPSW', ipsw)
		dbg_set_sys_reg('PSW', psw)
	else
		psw = dbg_get_sys_reg('PSW')
		psw = bit_and(psw, 0xfffff7ff)  -- disable HSS
		dbg_set_sys_reg('PSW', psw)
	end
	aice.set_cmd_verbose(v)
end

function dbg_wait_crst()
	local v = aice.set_cmd_verbose(0)
	local interval = 100
	local dbger = 0
	local retry = 0
	local RETRY_COUNT = 50	-- 50*100 = 5000 ms
	while dbger == 0 do
		dbger = dbg_get_dbger()
		dbger = bit_and(dbger, 0x4)
		aice.sleep(interval)
		retry = retry + 1
		if (retry > RETRY_COUNT) then
			return
		end
	end
	dbg_set_dbger(0x4) -- clear CRST
	aice.set_cmd_verbose(v)
end

function dbg_restart_target(...) 
	local dim_base, init_sys_rst_delay = ...
	local v = aice.set_cmd_verbose(0)
	local soft_restarted = false

	if not pcall(dbg_hard_restart, dim_base, init_sys_rst_delay) then
		aice.cmd_clear_timeout()
		dbg_soft_restart(dim_base, init_sys_rst_delay)
		soft_restarted = true
	end

	local result, ivb = pcall(dbg_get_sys_reg, "IVB")
	if (not result and soft_restarted) then
		error(string.format("failed to get IVB after dbg_soft_restart(): %s\n", ivb), 2)
	end
	if (not result and not soft_restarted) then
		aice.cmd_clear_timeout()
		dbg_soft_restart(dim_base, init_sys_rst_delay)
		ivb = dbg_get_sys_reg("IVB")
	end

	ivb = bit_lshift(bit_rshift(ivb, 16), 16)

	local pc = dbg_get_pc()
	if (pc ~= ivb) then
		dbg_set_pc(ivb)
	end

	local psw = dbg_get_sys_reg("PSW")
	psw = bit_or(psw, 0x1f)
	psw = bit_xor(psw, 0x15)	-- gie = 0, intl = 1
	dbg_set_sys_reg("PSW", psw)

	psw = bit_or(psw, 0x400)
	psw = bit_xor(psw, 0x400)	-- dex = 0
	dbg_set_sys_reg("IPSW", psw)
	
	dbg_set_sys_reg("ITYPE", 0)

	local msc_cfg = dbg_get_sys_reg('MSC_CFG')
	local intlc = bit_test(msc_cfg, 0x1000)
	if (not intlc) then
		dbg_set_sys_reg("P_ITYPE", 0)
	end
	aice.set_cmd_verbose(v)
end

function dbg_hard_restart(...) 
	local dim_base, init_sys_rst_delay = ...
	local v = aice.set_cmd_verbose(0)

	local aice_box = is_aice_box()
	local fw_ver = aice.cmd_read_ctrl(0x3) -- firmware version
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end

	local aice_ctrl4 = aice.cmd_read_ctrl(0x4)
	if (aice_cmd_timeout) then
		error("connectivity to AICE broken", 2)
	end

	if (aice_box) then
		-- assume 7-pin first
		if (bit_test(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)) then

			-- clear AICE_CTRL4_NODBGI_MASK
			aice_ctrl4 = bit_or(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)
			aice_ctrl4 = bit_xor(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)
			aice.cmd_write_ctrl(0x4, aice_ctrl4)
		end
	end

	local sysrst_delay = init_sys_rst_delay

	if (dim_base == nil) then
		dim_base = 0xfff00000
	end
	if (sysrst_delay == nil) then
		if (aice_mcu) then
			sysrst_delay = 600
		else
			sysrst_delay = 700
		end
	end

	local RETRY_COUNT = 2
	local done, retry, warning_shown
	retry = 0
	done = true
	warning_shown = false

	aice.cmd_write_edmsr(s('DIMBR'), dim_base)
	if (aice_cmd_timeout) then
		done = false
		aice.cmd_clear_timeout()
	end

	while not done do
		-- set DIMBR and train sysrst_delay at the same time

		aice.cmd_sys_reset()
		-- do not use dbg_wait_crst() here to shorten reset delay
		aice.sleep(sysrst_delay)

		aice.cmd_write_edmsr(s('DIMBR'), dim_base)
		if (aice_cmd_timeout) then
			aice.cmd_clear_timeout()
			if (not warning_shown and sysrst_delay > 800) then
				warning_shown = true
				printf("WARNING: target is not responding...trying sys_reset with more delay\n")
			end

			if (sysrst_delay < 1200) then
				sysrst_delay = sysrst_delay + sysrst_delay / 2 + 11
			elseif (sysrst_delay < 3000) then
				sysrst_delay = sysrst_delay + 501
			end
			if (sysrst_delay > 3000) then
				error("can not establish connection with DUT", 2)
			end
		else
			done = true
		end
	end

	dbg_set_dbger(0x7) -- clear CRST, DPEX and DEX
	if (aice_cmd_timeout) then
		aice.cmd_clear_timeout()
	end

	aice.cmd_restart()
	if (aice_cmd_timeout and not aice_mcu) then
		if (fw_ver >= 6) then
			aice.cmd_clear_timeout()

			local edm_cfg = aice.cmd_read_edmsr(s('EDM_CFG'))
			if (not bit_test(edm_cfg, 0xf0000000)) then
				edm_cfg = bit_rshift(edm_cfg, 16)
				if (edm_cfg <= 0x0050) then
					-- EDM version <= 0x0050 do not function correctly with 5-pin settings
					error("hardware restart failed. please try dbg_soft_restart(...)\n", 2)
				end
			end
			-- retry with 5-pin (no dbgi_n pin)
			aice_ctrl4 = bit_or(aice_ctrl4, AICE_CTRL4_NODBGI_MASK)
			aice.cmd_write_ctrl(0x4, aice_ctrl4)

			aice.cmd_restart()
		end
	end
	if (aice_cmd_timeout) then
		error(string.format("connect to AICE failed"), 2)
	end
	aice.sleep(sysrst_delay)
	local dbger = dbg_get_dbger()
	if (not bit_test(dbger, 0x4)) then
		error("hardware restart unsuccessful: system reset didn't happen\n", 2)
	end

	local edm_cfg = aice.cmd_read_edmsr(s('EDM_CTL'))
	if (not bit_test(edm_cfg, 0x80000000)) then
		error("hardware restart unsuccessful: debug memory is not activated", 2)
	end

	aice.set_cmd_verbose(v)
end

function dbg_soft_restart(...) 
	local dim_base, init_sys_rst_delay = ...
	local dummy_cmd = {}
	local v = aice.set_cmd_verbose(0)
	local sysrst_delay = init_sys_rst_delay

	local aice_box = is_aice_box()

	if (dim_base == nil) then
		dim_base = 0xfff00000
	end
	if (sysrst_delay == nil) then
		if (not aice_box) then
			sysrst_delay = 3
		else
			sysrst_delay = 6
		end
	end

	local RETRY_COUNT = 2
	local done, retry
	retry = 0
	done = false

	-- set DIMBR and train sysrst_delay at the same time

	while not done do
		aice.cmd_sys_reset()
		-- do not use dbg_wait_crst() here to shorten reset delay
		aice.sleep(sysrst_delay)

		aice.cmd_write_edmsr(s('DIMBR'), dim_base)
		if (aice_cmd_timeout) then
			aice.cmd_clear_timeout()
			if (sysrst_delay < 25) then
				sysrst_delay = sysrst_delay + sysrst_delay / 2 + 1
			elseif (sysrst_delay < 250) then
				sysrst_delay = 2*sysrst_delay + 1
			elseif (sysrst_delay < 1900) then
				sysrst_delay = 2*sysrst_delay + 11
			elseif (sysrst_delay < 5000) then
				sysrst_delay = sysrst_delay + 501
			end
			if (sysrst_delay > 5000) then
				error("can not establish connection with DUT", 2)
			end
		else
			done = true
		end
	end

	sysrst_delay = sysrst_delay + 1

	RETRY_COUNT = 2
	done = false
	aice.set_usb_verbose(0)
	while not done do
		aice.cmd_write_ctrl(0x01, 0x3) -- sysrst & trst
		if (aice_cmd_timeout) then
			error("connect to AICE failed", 2)
		end
		aice.sleep(sysrst_delay)
		aice.cmd_write_edmsr(s('DIMBR'), dim_base)
		if (aice_cmd_timeout) then
			error("failed", 2)
		end
		aice.set_pack_mode(1)
		aice.cmd_write_edmsr(s('EDM_CTL'), 0x80000000)
		aice.cmd_write_misc(m('EDM_CMDR'), 0x0) -- trigger the debug exception
		aice.set_pack_mode(0)

		-- more setup
		aice.cmd_write_misc(m('DIMIR'), 0x0)
		if (aice_cmd_timeout) then
			retry = retry + 1
			if (retry >= RETRY_COUNT) then
				error("connect to DUT failed", 2)
				os.exit(1)
			end
			aice.cmd_clear_timeout()
		else
			done = true
		end
	end
	aice.set_usb_verbose(0)
	
	dbg_set_dbger(0x2) -- clear DPED
	aice.set_cmd_verbose(v)
end

--                 1: ! init r1            ! write index
-- 64170002              mfsr $r1, $EDM_DTR
-- 40000009              nop
-- 40000009              nop
-- 48fffffa              j 1b
function dbg_set_gpr(reg_idx, value)
	local v = aice.set_cmd_verbose(0)
	local reg

	dbg_set_dtr(value)

	local w = {0x64070002, 0x40000009, 0x40000009, 0x48fffffa}
	reg_idx = bit_and(reg_idx, 0x1f)
	w[1] = bit_or(w[1], bit_lshift(reg_idx, 20))
	dbg_execute(w)

	aice.set_cmd_verbose(v)
end

--          	1: ! backup r1
-- 64170003              mtsr $r1, $EDM_DTR
-- 64000008              dsb
-- 40000009              nop
-- 48fffffa              j 1b
function dbg_get_gpr(reg_idx)
	local v = aice.set_cmd_verbose(0)
	local reg

	local w = {0x64070003, 0x64000008, 0x40000009, 0x48fffffa}
	reg_idx = bit_and(reg_idx, 0x1f)
	w[1] = bit_or(w[1], bit_lshift(reg_idx, 20))
	dbg_execute(w)

	reg = dbg_get_dtr()
	aice.set_cmd_verbose(v)

	return reg
end

-- 14000000              swi  $r0, [$r0]
-- 40000009              nop
-- 40000009              nop
-- 48fffffa              j 1b
function dbg_swi(reg_rt, reg_ra, offset)
	local v = aice.set_cmd_verbose(0)

	local w = {0x14000000, 0x40000009, 0x40000009, 0x48fffffa}
	reg_rt = bit_and(reg_rt, 0x1f)
	reg_ra = bit_and(reg_ra, 0x1f)

	offset = bit_and(bit_arshift(offset, 2), 0x7fff)
	w[1] = bit_or(w[1], bit_lshift(reg_rt, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[1] = bit_or(w[1], offset)
	dbg_execute(w)

	aice.set_cmd_verbose(v)
end

-- 04000000              lwi  $r0, [$r0]
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_lwi(reg_rt, reg_ra, offset)
	local v = aice.set_cmd_verbose(0)

	local w = {0x04000000, 0x64070003, 0x64000008, 0x48fffffa}
	reg_rt = bit_and(reg_rt, 0x1f)
	reg_ra = bit_and(reg_ra, 0x1f)
	offset = bit_and(bit_arshift(offset, 2), 0x7fff)
	w[1] = bit_or(w[1], bit_lshift(reg_rt, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[1] = bit_or(w[1], offset)

	w[2] = bit_or(w[2], bit_lshift(reg_rt, 20))
	dbg_execute(w)

	local rt_value = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error(string.format("failed to get r%0d", reg_rt), 2)
	end
	aice.set_cmd_verbose(v)

	return rt_value
end

-- 10000000              sbi  $r0, [$r0]
-- 40000009              nop
-- 40000009              nop
-- 48fffffa              j 1b
function dbg_sbi(reg_rt, reg_ra, offset)
	local v = aice.set_cmd_verbose(0)

	local w = {0x10000000, 0x40000009, 0x40000009, 0x48fffffa}
	reg_rt = bit_and(reg_rt, 0x1f)
	reg_ra = bit_and(reg_ra, 0x1f)

	offset = bit_and(offset, 0x7fff)
	w[1] = bit_or(w[1], bit_lshift(reg_rt, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[1] = bit_or(w[1], offset)
	dbg_execute(w)

	aice.set_cmd_verbose(v)
end

-- 00000000              lbi  $r0, [$r0]
-- 64070003              mtsr    $r0, $EDM_DTR
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_lbi(reg_rt, reg_ra, offset)
	local v = aice.set_cmd_verbose(0)

	local w = {0x00000000, 0x64070003, 0x64000008, 0x48fffffa}
	reg_rt = bit_and(reg_rt, 0x1f)
	reg_ra = bit_and(reg_ra, 0x1f)
	offset = bit_and(offset, 0x7fff)
	w[1] = bit_or(w[1], bit_lshift(reg_rt, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[1] = bit_or(w[1], offset)

	w[2] = bit_or(w[2], bit_lshift(reg_rt, 20))
	dbg_execute(w)

	local rt_value = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error(string.format("failed to get r%0d", reg_rt), 2)
	end
	aice.set_cmd_verbose(v)

	return rt_value
end

-- 64070002              mfsr   $r0, $EDM_DTR
-- 6400000d              isync  $r0
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_isync(va)
	local v = aice.set_cmd_verbose(0)
	
	dbg_set_dtr(va)

	local w = {0x64070002, 0x6400000d, 0x64000008, 0x48fffffa}
	
	dbg_execute(w)
	
	aice.set_cmd_verbose(v)
end

-- 64000001              cctl  $rt, $ra, subtype
-- 64070003              mtsr  $rt, $EDM_DTR
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_cctl(...)
	local reg_rt, reg_ra, subtype, alevel = ...

	local v = aice.set_cmd_verbose(0)

	local w = {0x64000001, 0x64070003, 0x64000008, 0x48fffffa}
	reg_rt = bit_and(reg_rt, 0x1f)
	reg_ra = bit_and(reg_ra, 0x1f)
	subtype = bit_lshift(bit_and(subtype, 0x1f), 5)
	w[1] = bit_or(w[1], bit_lshift(reg_rt, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[1] = bit_or(w[1], subtype)

	w[2] = bit_or(w[2], bit_lshift(reg_rt, 20))
	dbg_execute(w)

	local rt_value = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error(string.format("failed to get r%0d", reg_rt), 2)
	end
	aice.set_cmd_verbose(v)

	return rt_value
end

-- 64000001              cctl  $ra, $ra, subtype
-- 40000000              add   $ra, $ra, $rb
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_cctl_add(...)
	local reg_ra, reg_rb, subtype, alevel = ...

	local v = aice.set_cmd_verbose(0)

	local w = {0x64000001, 0x40000000, 0x64000008, 0x48fffffa}
	reg_ra = bit_and(reg_ra, 0x1f)
	reg_rb = bit_and(reg_rb, 0x1f)
	subtype = bit_lshift(bit_and(subtype, 0x1f), 5)
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[1] = bit_or(w[1], subtype)

	w[2] = bit_or(w[2], bit_lshift(reg_ra, 20))
	w[2] = bit_or(w[2], bit_lshift(reg_ra, 15))
	w[2] = bit_or(w[2], bit_lshift(reg_rb, 10))
	dbg_execute(w)

	aice.set_cmd_verbose(v)
end

-- 6400000e		 tlbop $r0, TargetRead
-- 64000008              nop
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_tlbop_target_read(...)
	local reg_ra = ...
	local v = aice.set_cmd_verbose(0)

	if (reg_ra == nil) then
		reg_ra = 0
	end
	reg_ra = bit_and(reg_ra, 0x1f)

	local w = {0x6400000e, 0x64000008, 0x64000008, 0x48fffffa}

	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	dbg_execute(w)

	aice.set_cmd_verbose(v)
end

-- 640000ae		 tlbop $r0, $r0, Probe
-- 64070003              mtsr  $r0, $EDM_DTR
-- 64000008              nop
-- 48fffffa              j 1b
function dbg_tlbop_probe(...)
	local reg_rt, reg_ra = ...
	local v = aice.set_cmd_verbose(0)

	if (reg_rt == nil) then
		reg_rt = 0
	end
	if (reg_ra == nil) then
		reg_ra = reg_rt
	end
	reg_ra = bit_and(reg_ra, 0x1f)

	local w = {0x640000ae, 0x64070003, 0x64000008, 0x48fffffa}

	w[1] = bit_or(w[1], bit_lshift(reg_rt, 20))
	w[1] = bit_or(w[1], bit_lshift(reg_ra, 15))
	w[2] = bit_or(w[2], bit_lshift(reg_rt, 20))
	dbg_execute(w)

	local rt_value = dbg_get_dtr()
	if (aice_cmd_timeout) then
		error(string.format("failed to get r%0d", reg_rt), 2)
	end
	aice.set_cmd_verbose(v)

	return rt_value
end

function dbg_load_file(addr, filename)
	local v = aice.set_cmd_verbose(0)
	local image = aice.load_image(filename)
	for i=1, #image do
		local word = image[i]
		local _addr = addr+(i-1)*4
		aice.cmd_write_word(_addr, word)
		if (aice_cmd_timeout) then
			break
		end
	end
	aice.set_cmd_verbose(v)
end

-- #chunk must be <= 256
function dbg_fastwrite_mem(addr, chunk)
	local v = aice.set_cmd_verbose(0)
	-- align to word
	local a = bit_rshift(addr, 2)
	a = bit_lshift(a, 2)
	a = bit_or(a, 1)	-- bit 0 of SBAR == 1 indicates this is a fast memory write

	dbg_set_edm_reg('SBAR', a)
	if (aice_cmd_timeout) then
		printf("aice timeout writing %08x to SBAR\n", a)
		return
	end
	aice.cmd_fastwrite_mem(chunk)
	aice.set_cmd_verbose(v)
end

-- n must be <= 256
function dbg_fastread_mem(addr, n)
	local v = aice.set_cmd_verbose(0)
	-- align to word
	local a = bit_rshift(addr, 2)
	a = bit_lshift(a, 2)
	a = bit_or(a, 0)	-- bit 0 of SBAR == 0 indicates this is a fast memory read

	dbg_set_edm_reg('SBAR', a)
	if (aice_cmd_timeout) then
		printf("aice timeout writing %08x to SBAR\n", a)
		return
	end

	local mem = aice.cmd_fastread_mem(n)
	aice.set_cmd_verbose(v)

	return mem
end

function dbg_fast_load_file(addr, filename)
	local image = aice.image(filename)
	local k
	k = 0
	while (k < #image) do
		local chunk = {}
		for i=1,256 do
			k = k + 1
			if (k > #image) then
				break
			end
			chunk[i] = image[k]
		end
		do_fastwrite_mem(addr, chunk)
		addr = addr + (#chunk*4)
		if (aice_cmd_timeout) then
			break
		end
	end
end

function dbg_block_dump(addr, size)
	local v = aice.set_cmd_verbose(0)
	for a = addr, addr+size-1, 4 do
		w = aice.cmd_read_word(a)
		if (aice_cmd_timeout) then
			break
		end
		printf("[%08x] = 0x%08x\n", a, w)
	end
	aice.set_cmd_verbose(v)
end

function opt_clear_timeout()
	local v = aice.set_cmd_verbose(0)
	local reg = 0x10 -- RDY = 1

	if not aice_cmd_timeout then
		reg = aice.cmd_read_ctrl(0x00)	-- only valid for AICE, not AICE_MCU
		if (aice_cmd_timeout) then
			error("USB connection or icebox broken?!\n", 2)
		end

		if (reg == nil) then
			reg = 0
		else
			reg = bit_and(reg, 0x10)
		end
	else
		reg = 0x0
	end

	if (reg == 0) then
		aice.cmd_clear_timeout()
	end

	if (aice_cmd_timeout) then
		error("USB connection or icebox broken?!\n", 2)
	end

	aice.set_cmd_verbose(v)
end

function clear_timeout()
	local v = aice.set_cmd_verbose(0)

	aice.cmd_clear_timeout()
	if (aice_cmd_timeout) then
		error("USB connection or icebox broken?!\n", 2)
	end

	-- workaroud for AICE-MCU v1.3.2: AICE-MCU would enable fast mode after clear timeout
	if (not _G['sdp_fast_mode']) then
		disable_sdp_fast_mode()
	end
	aice.set_cmd_verbose(v)
end

function check_timeout(format, ...)
	if (aice_cmd_timeout) then
		local msg = string.format(format, ...)
		error(msg, 2)
	end
end

-- report caller location instead of the location of check_timeout() call.
function lib_check_timeout(format, ...)
	if (aice_cmd_timeout) then
		local msg = string.format(format, ...)
		error(msg, 3)
	end
end

cmd_read_mem = cmd_read_word
cmd_write_mem = cmd_write_word

function dbg_set_watch_point(n, addr, mask)
	local bpa_tag = "BPA" .. n
	local bpam_tag = "BPAM" .. n
	local bpc_tag = "BPC" .. n
	dbg_set_edm_reg(bpa_tag, addr)
	dbg_set_edm_reg(bpam_tag, mask)
	 -- 0_0000_1101 = 0x00d
	dbg_set_edm_reg(bpc_tag, 0xd)
end

function dbg_set_break_point(n, addr, mask)
	local bpa_tag = "BPA" .. n
	local bpam_tag = "BPAM" .. n
	local bpc_tag = "BPC" .. n
	dbg_set_edm_reg(bpa_tag, addr)
	dbg_set_edm_reg(bpam_tag, mask)
	 -- 0_0000_1010 = 0x00a
	dbg_set_edm_reg(bpc_tag, 0xa)
end

function dbg_clr_break_point(n)
	local bpc_tag = "BPC" .. n
	dbg_set_edm_reg(bpc_tag, 0x0)
end

dbg_clr_watch_point = dbg_clr_break_point

function decode_tck_option(opt_clock)
	local tck_freq
	local tck_name
	if (opt_clock == nil) then
		tck_freq = TCK_48MHz
		tck_name = "24MHz"
	else 
		opt_clock = opt_clock + 0
		if (opt_clock == 48) then
			tck_freq = TCK_48MHz
			tck_name = "48MHz"
		elseif (opt_clock == 24) then
			tck_freq = TCK_24MHz
			tck_name = "24MHz"
		elseif (opt_clock == 12) then
			tck_freq = TCK_12MHz
			tck_name = "12MHz"
		elseif (opt_clock == 6) then
			tck_freq = TCK_6MHz
			tck_name = "6MHz"
		elseif (opt_clock == 3) then
			tck_freq = TCK_3MHz
			tck_name = "3MHz"
		elseif (opt_clock == 1500) then
			tck_freq = TCK_1500KHz
			tck_name = "1.5MHz"
		elseif (opt_clock == 750) then
			tck_freq = TCK_750KHz
			tck_name = "750KHz"
		elseif (opt_clock == 30) then
			tck_freq = TCK_30MHz
			tck_name = "30MHz"
		elseif (opt_clock == 15) then
			tck_freq = TCK_15MHz
			tck_name = "15MHz"
		elseif (opt_clock == 7) then
			tck_freq = TCK_7MHz
			tck_name = "7MHz"
		elseif (opt_clock == 3750) then
			tck_freq = TCK_3750KHz
			tck_name = "3750KHz"
		end
	end

	return tck_freq, tck_name
end

-- usage:
-- show all wanted registers:
-- 	dbg_show_regs({0, 1, 2, 3, 4, 5, 6, 15})
-- show only updated registers since last call:
-- 	dbg_show_regs({0, 1, 2, 3, 4, 5, 6, 15}, gpr)
function dbg_show_regs(...)
	local wanted_regs, last_values = ...
	local reg_num

	if (wanted_regs == nil) then
		wanted_regs = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}
	end

	if (last_values == nil) then
		last_values = {}
	end

	for _,reg_num in ipairs(wanted_regs) do
		local v = dbg_get_gpr(reg_num)
		if (last_values[reg_num] == nil or last_values[reg_num] ~= v) then
			printf(" r%0d=%08x", reg_num, v)
		end
		last_values[reg_num] = v
	end
        printf("\n")
end
